--Loop Control v1.01 by Aftab Hussain aka afta8 - fathand@gmail.com - 22nd April 2014

--[[============================================================================
main.lua
============================================================================]]--

-- Placeholder for the dialog
local dialog = nil

-- Placeholder to expose the ViewBuilder outside the show_dialog() function
local vb = nil

-- Reload the script whenever this file is saved. 
-- Additionally, execute the attached function.
_AUTO_RELOAD_DEBUG = function()
  
end

-- Read from the manifest.xml file.
class "RenoiseScriptingTool" (renoise.Document.DocumentNode)
  function RenoiseScriptingTool:__init()    
    renoise.Document.DocumentNode.__init(self) 
    self:add_property("Name", "Untitled Tool")
    self:add_property("Id", "Unknown Id")
  end

local manifest = RenoiseScriptingTool()
local ok,err = manifest:load_from("manifest.xml")
local tool_name = manifest:property("Name").value
local tool_id = manifest:property("Id").value


--------------------------------------------------------------------------------
-- Helper functions
--------------------------------------------------------------------------------

-- Math functions
local min = math.min
local max = math.max
local floor = math.floor
local log = math.log
local pow = math.pow

-- Interpolate
local function interpolate(pos, min, max) -- Input must be 0 to 1
  return pos*(max-min)+min
end

-- Reverse Interpolate
local function rev_interpolate(output, min, max)
  return (output-min)/(max-min)
end

-- Clamp values
local function clamp_value(input, min_val, max_val)
  return min(max(input, min_val), max_val)
end

-- Integer rounding
local function round(num) 
  return floor(num+0.5) 
end

-- Round to a step value
function round_step(num, step)
  return floor((num/step) + 0.5)*step
end

-- log2(x)
local function log2(x) 
  return log(x) / log(2) 
end

-- Frequency to MIDI
local function freq2midi(x)
  return 69+(12*log2(x/440))
end

-- MIDI to Frequency
local function midi2freq(x)
  return pow(2,((x-69)/12))*440
end

-- Frequency from sample size
local function smp2freq(sample_size, sample_rate)
  return sample_rate/sample_size
end

-- Sample size from frequency
local function freq2smp(frequency, sample_rate)
  return sample_rate/frequency
end  

-- Notifier handlers
local notifier = {}
  function notifier.add(observable, n_function)
    if not observable:has_notifier(n_function) then
      observable:add_notifier(n_function)
    end
  end

  function notifier.remove(observable, n_function)
    if observable:has_notifier(n_function) then
      observable:remove_notifier(n_function)
    end
  end

-- Read xml preset data and apply to given device
local function apply_preset_data(xml_data, device)
  local infile = io.open(xml_data, "rb")
  local indata = infile:read("*all")
  device.active_preset_data = indata
  infile:close()
end


--------------------------------------------------------------------------------
-- Main functions
--------------------------------------------------------------------------------

-- Set up preferences
local options = renoise.Document.create("ScriptingToolPreferences") {
  always_on = true
}
renoise.tool().preferences = options


-- Global table of settings
local t = {}

-- Get sample size
local function smp_size(sample)
  return sample.sample_buffer.number_of_frames
end

-- Get sample rate
local function smp_rate(sample)
  return sample.sample_buffer.sample_rate
end


-- Table of snap mode functions
local snap_modes = {}
-- Free mode
snap_modes[1] = { func = function (r, x)
                           return x
                         end,
                 
                  param = function (dial)
                            return dial
                          end
                }

-- Return nearest beat division size
snap_modes[2] = { func = function (r, sample_size, beat_div)
                           return round_step(sample_size, smp_size(t[r].smp)/beat_div)
                         end,
                         
                  param = function (dial)
                            local values = {2,4,8,16,32,64}
                            local dial_pos = floor(interpolate(dial*0.01, 1, 6.9))
                            return values[dial_pos]
                          end
                }

-- Given current sample size return nearest sample size that corresponds to a note
snap_modes[3] = { func = function (r, sample_size)
                           -- Get closest note from given sample size
                           local note = round(freq2midi(smp2freq(sample_size, smp_rate(t[r].smp))))
                           -- Return sample size for nearest note
                           return round(freq2smp((midi2freq(note)), smp_rate(t[r].smp)))
                         end,
                 
                   param = function (dial)
                             return dial
                           end                                
                }


-- Get loop size
local function loop_size(r)
  local loop_size = interpolate(t[r].loop_size_dial.value*0.01, 1, smp_size(t[r].smp))
  local snap_mode = floor(interpolate(t[r].loop_size_mode.value*0.01, 1, 3.9))
  return snap_modes[snap_mode].func(r, loop_size, snap_modes[snap_mode].param(t[r].loop_size_param.value))
end


-- Get loop end position
local function loop_end(r)
  local loop_end = interpolate(t[r].loop_end_dial.value*0.01, 1, smp_size(t[r].smp))
  local snap_mode = floor(interpolate(t[r].loop_end_mode.value*0.01, 1, 2.9))
  return snap_modes[snap_mode].func(r, loop_end, snap_modes[snap_mode].param(t[r].loop_end_param.value))
end



-- Function to create device chain
local function init_device_chain(instrument, device_chain_pos)
  local device_chain = instrument:insert_sample_device_chain_at(device_chain_pos)
  device_chain.name = "~Loop Control"
  return device_chain
end


-- Function to create doofer device
local function init_doofer(device_chain, sample_index)
  local device_pos = clamp_value(sample_index+1, 2, #device_chain.devices+1)
  local doofer = device_chain:insert_device_at("Audio/Effects/Native/Doofer", device_pos)
  apply_preset_data("doofer.xml", doofer)
  doofer.display_name = tostring(sample_index)
  return doofer
end



-- Return the ~Loop Control device chain if it already exists
local function find_device_chain(instrument)
  local device_chain = false
  if #instrument.sample_device_chains > 0 then
    for dc = 1, #instrument.sample_device_chains do  
      if instrument.sample_device_chains[dc].name == "~Loop Control" then 
        device_chain = instrument.sample_device_chains[dc]       
      end  
    end
  end
  return device_chain
end


-- Return the ~Loop control doofer if sample is controlled
local function find_doofer(device_chain, sample_index)
  local doofer = false  
  if #device_chain.devices > 1 then
    for d = 2, #device_chain.devices do
      if device_chain.devices[d].display_name == tostring(sample_index) then
        doofer = device_chain.devices[d]
      end
    end
  end
  return doofer
end


-- Set up loop control on given sample
local function setup(r, ins_index, smp_index, device_chain, device)

  t[r] = {}
  
  -- Set instrument and sample
  t[r].ins = renoise.song().instruments[ins_index]
  t[r].ins_index = ins_index
  t[r].smp = t[r].ins.samples[smp_index]
  t[r].smp_index = smp_index
    
  -- Set device chain
  if device_chain == false then
    t[r].device_chain = init_device_chain(t[r].ins, 1)
  else
    t[r].device_chain = device_chain
  end
  
  --Set instrument fx device
  if device == false then
    t[r].device = init_doofer(t[r].device_chain, t[r].smp_index)
  else
    t[r].device = device
  end
  
  -- Turn loop on and get existing loop settings
  if t[r].smp.loop_mode < 2 then t[r].smp.loop_mode = 2 end
  
  -- Set observed loop start and end macros
  t[r].loop_size_dial = t[r].device.parameters[1]
  t[r].loop_size_mode = t[r].device.parameters[2]
  t[r].loop_size_param = t[r].device.parameters[3]

  t[r].loop_end_dial = t[r].device.parameters[4]
  t[r].loop_end_mode = t[r].device.parameters[5]
  t[r].loop_end_param = t[r].device.parameters[6]

  -- Set loop loop start and end dials based on existing loop setings
  t[r].loop_end_dial.value = clamp_value(rev_interpolate(t[r].smp.loop_end, 1, smp_size(t[r].smp))*100, 1, 100)
  t[r].loop_size_dial.value = clamp_value(rev_interpolate(t[r].smp.loop_end - t[r].smp.loop_start, 1, smp_size(t[r].smp))*100, 1, 100)
       
  -- Set up observables for notifier functions
  t[r].loop_end_observable = t[r].loop_end_dial.value_observable
  t[r].loop_size_observable = t[r].loop_size_dial.value_observable

  -- Create notifier function
  t[r].n_function = function() 
                      -- Set the loop end position
                      t[r].smp.loop_end = clamp_value(loop_end(r), 1, smp_size(t[r].smp)) 
                      -- Set the loop start position
                      t[r].smp.loop_start = clamp_value((loop_end(r)-loop_size(r)), 1, smp_size(t[r].smp))                  
                    end
                                       
  -- Add notifiers
  notifier.add(t[r].loop_size_observable, t[r].n_function)
  notifier.add(t[r].loop_end_observable, t[r].n_function)

end


-- Add loop control to currently selected sample
local function add_loop_control()

  local ins_index = renoise.song().selected_instrument_index
  local smp_index = renoise.song().selected_sample_index
  local device_chain = find_device_chain(renoise.song().instruments[ins_index])
  local device = false
  if device_chain ~= false then 
    device = find_doofer(device_chain, smp_index)
  end

  local r = #t + 1
  setup(r, ins_index, smp_index, device_chain, device)
    
end


-- Song opening function, scan instruments and add loop control where relevant
local function on_new_song()
  -- Initialise settings
  t = {}
  
  if options.always_on.value then 
  
    -- Cycle instruments 
    for ins_index = 1, #renoise.song().instruments do 
      -- Check for a loop control fx chain
      local device_chain = find_device_chain(renoise.song().instruments[ins_index])
      if device_chain ~= false then -- Device chain detected
        -- Cycle through samples     
        for smp_index = 1, #renoise.song().instruments[ins_index].samples do         
          -- Check if there is a doofer for the sample
          local device = find_doofer(device_chain, smp_index)
          if device ~= false then -- doofer detected
            -- Set up loop control on sample
            local r = #t + 1
            setup(r, ins_index, smp_index, device_chain, device)    
          end -- if device        
        end -- smp loop
      end -- if device chains
    end -- ins loop

  end -- Options

end


-- Remove all notifiers when tool is deactivated in menu
local function remove_all_notifiers()
  for r = 1, #t do
    notifier.remove(t[r].loop_size_observable, t[r].n_function)
    notifier.remove(t[r].loop_end_observable, t[r].n_function)
  end
end


--------------------------------------------------------------------------------
-- Song Opening functions
--------------------------------------------------------------------------------

-- Set up observables
local new_doc_observable = renoise.tool().app_new_document_observable

-- Add the notifiers (always running in background)
notifier.add(new_doc_observable, on_new_song)


--------------------------------------------------------------------------------
-- Menu entries
--------------------------------------------------------------------------------

local function background_state() 
  return options.always_on.value 
end

local function background_enable() 
  options.always_on.value = not options.always_on.value
  remove_all_notifiers()
  on_new_song()
end


renoise.tool():add_menu_entry {
  name = "Main Menu:Tools:"..tool_name..":Always On",
  invoke = background_enable,
  selected = background_state
}

renoise.tool():add_menu_entry {
  name = "Sample Navigator:Add Loop Control",
  invoke = add_loop_control
}

--------------------------------------------------------------------------------
-- Key Binding
--------------------------------------------------------------------------------

renoise.tool():add_keybinding {
  name = "Global:Tools:" .. tool_name.."...",
  invoke = add_loop_control
}

--------------------------------------------------------------------------------
-- MIDI Mapping
--------------------------------------------------------------------------------

--[[
renoise.tool():add_midi_mapping {
  name = tool_id..":Show Dialog...",
  invoke = show_dialog
}
--]]
